/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*-
 
 * The contents of this file are subject to the Netscape Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 *
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s):
 *   Peter Hartshorn <peter@igelaus.com.au>
 *   B.J. Rossiter <bj@igelaus.com.au>
 *   Matthew Wlazlo <matthew.wlazlo@tuxia.com>
 */

#include "xp_core.h"
#include "nsQuickSort.h" 
#include "nsFontMetricsNanoX.h"
#include "nspr.h"
#include "nsIPref.h"
#include "nsIServiceManager.h"

static NS_DEFINE_IID(kIFontMetricsIID, NS_IFONT_METRICS_IID);
static NS_DEFINE_CID(kPrefCID, NS_PREF_CID);

static GR_FONTLIST **gFontList;
static int gNumFonts;

static PRBool
IsASCIIFontName(const nsString& aName)
{
  PRUint32 len = aName.Length();
  const PRUnichar* str = aName.get();
  for (PRUint32 i = 0; i < len; i++) {
    /*
     * X font names are printable ASCII, ignore others (for now)
     */
    if ((str[i] < 0x20) || (str[i] > 0x7E)) {
      return PR_FALSE;
    }
  }

  return PR_TRUE;
}

static PRBool FontEnumCallback(const nsString& aFamily, PRBool aGeneric,
                               void *aData)
{

  if (!IsASCIIFontName(aFamily))
    return PR_TRUE; // skip and continue

  nsCAutoString name;
  name.AssignWithConversion(aFamily.get());
  name.ToLowerCase();

  nsFontMetricsNanoX *metrics = (nsFontMetricsNanoX*)aData;
  metrics->mFonts.AppendCString(name);
  metrics->mFontIsGeneric.AppendElement((void*)aGeneric);

  if (aGeneric) {
    metrics->mGeneric = metrics->mFonts.CStringAt(metrics->mFonts.Count()-1);
    return PR_FALSE;
  }
  return PR_TRUE;
}

nsFontNanoX::nsFontNanoX()
{
  mFont = 0;
  //mName = nsnull;
  mHeight = 0;
}

nsFontNanoX::~nsFontNanoX()
{
  if (mFont)
    GrDestroyFont(mFont);
}

void nsFontNanoX::LoadFont()
{
	if(gFontList == NULL)
		GrGetFontList(&gFontList, &gNumFonts);
	
	for (int i = 0; i < gNumFonts; i++) {
		if (!strcmp(mName.get(), gFontList[i]->ttname)) {
			mFont=GrCreateFont((GR_CHAR*)gFontList[i]->mwname,mHeight,NULL);
			GrGetFontInfo(mFont, &fInfo);
			return;
		}
	}
  //mFont = GrCreateFont((GR_CHAR*)"arial", 12, NULL); 
  mFont = GrCreateFont((GR_CHAR*)"Terminal", mHeight, NULL);
  GrGetFontInfo(mFont, &fInfo);
}

nsFontMetricsNanoX::nsFontMetricsNanoX()
{
  NS_INIT_REFCNT();
  mDeviceContext = nsnull;
  mFont = nsnull;
  mFontHandle = nsnull;
  mDefaultFont = nsnull;

  mHeight = 0;
  mAscent = 0;
  mDescent = 0;
  mLeading = 0;
  mMaxAscent = 0;
  mMaxDescent = 0;
  mMaxAdvance = 0;
  mXHeight = 0;
  mSuperscriptOffset = 0;
  mSubscriptOffset = 0;
  mStrikeoutSize = 0;
  mStrikeoutOffset = 0;
  mUnderlineSize = 0;
  mUnderlineOffset = 0;
}

nsFontMetricsNanoX::~nsFontMetricsNanoX()
{
  if (nsnull != mFont) {
    delete mFont;
    mFont = nsnull;
  }
  if (nsnull != mDefaultFont) {
    delete mDefaultFont;
    mDefaultFont = nsnull;
  }
}

NS_IMPL_ISUPPORTS(nsFontMetricsNanoX, kIFontMetricsIID)

NS_IMETHODIMP nsFontMetricsNanoX::Init(const nsFont& aFont, nsIAtom* aLangGroup,
  nsIDeviceContext* aContext)
{
  nsIPref *aPref;

  mFont = new nsFont(aFont);
  mLangGroup = aLangGroup;
  mDeviceContext = aContext;

  float app2dev;
  aContext->GetAppUnitsToDevUnits(app2dev);

  float textZoom = 1.0;
  aContext->GetTextZoom(textZoom);

//  mPixelSize = NSToIntRound(app2dev * textZoom * mFont->size);
//  mStretchIndex = 4; // normal
 // mStyleIndex = mFont->style;
  nsCAutoString fontName;

  mFont->EnumerateFamilies(FontEnumCallback, this);
  nsServiceManager::GetService(kPrefCID, NS_GET_IID(nsIPref),
                            (nsISupports**)&aPref);

  char *value = nsnull;

  if (mDefaultFont != nsnull)
    delete mDefaultFont;

  mDefaultFont = new nsFontNanoX();

  if (!mGeneric) {
    aPref->CopyCharPref("font.default", &value);
    if (value) {
      fontName = value;
      nsMemory::Free(value);
      value = nsnull;
    } else
      fontName = "serif";

    mGeneric = &fontName;
  }

  if (mLangGroup) {
    nsCAutoString name("font.name.");
    const PRUnichar *langGroup = nsnull;

    if (mGeneric->Equals("monospace"))
      name.Append("monospace.");
    else
      name.Append(*mGeneric);

    name.Append(char('.'));

    mLangGroup->GetUnicode(&langGroup);
    name.AppendWithConversion(langGroup);
    aPref->CopyCharPref(name, &value);
    if (value) {
      fontName = value;
      nsMemory::Free(value);
      value = nsnull;
    } else
      fontName = "serif";
  }
  mGeneric = &fontName;

  if (mFont->weight >= NS_FONT_WEIGHT_BOLD)
    mGeneric->Append(" Bold");
  if (mFont->style == NS_FONT_STYLE_ITALIC)
    mGeneric->Append(" Italic");

  mDefaultFont->mName = *mGeneric;
  mDefaultFont->mHeight = NSToIntRound(mFont->size * app2dev);
  mDefaultFont->LoadFont();
  mFontHandle = mDefaultFont->mFont;

  RealizeFont(aContext);

  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::Destroy()
{
  return NS_OK;
}


void nsFontMetricsNanoX::RealizeFont(nsIDeviceContext* aContext)
{
  float f;
  GR_FONT_INFO fInfo;

  mDeviceContext->GetDevUnitsToAppUnits(f);

  GrGetFontInfo(mFontHandle, &fInfo);
  fInfo = mDefaultFont->fInfo;
  mMaxAscent = nscoord(fInfo.baseline * f);
  mMaxDescent = nscoord((fInfo.height - fInfo.baseline) * f);

  mMaxHeight = nscoord(fInfo.height * f);
  mEmHeight = PR_MAX(1, nscoord(mDefaultFont->fInfo.height * f));

  mMaxAdvance = nscoord(fInfo.maxwidth * f);
  mXHeight = NSToCoordRound((float)fInfo.baseline * f * 0.56f);

  mSpaceWidth = NSToCoordRound((float)fInfo.widths[' '] * f);

  mUnderlineOffset = -NSToIntRound(PR_MAX(1,floor(0.1 * fInfo.height + 0.5)*f));
  mUnderlineSize = 1;
  mSuperscriptOffset = mXHeight;
  mSubscriptOffset = mXHeight;

  mStrikeoutOffset = NSToCoordRound(mXHeight / 2.0);
  mStrikeoutSize = mUnderlineSize;

  mLeading = 0;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetXHeight(nscoord& aResult)
{
  aResult = mXHeight;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetSuperscriptOffset(nscoord& aResult)
{
  aResult = mSuperscriptOffset;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetSubscriptOffset(nscoord& aResult)
{
  aResult = mSubscriptOffset;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetStrikeout(nscoord& aOffset, nscoord& aSize)
{
  aOffset = mStrikeoutOffset;
  aSize = mStrikeoutSize;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetUnderline(nscoord& aOffset, nscoord& aSize)
{
  aOffset = mUnderlineOffset;
  aSize = mUnderlineSize;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetHeight(nscoord &aHeight)
{
  aHeight = mMaxHeight;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetLeading(nscoord &aLeading)
{
  aLeading = mLeading;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetMaxAscent(nscoord &aAscent)
{
  aAscent = mMaxAscent;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetMaxDescent(nscoord &aDescent)
{
  aDescent = mMaxDescent;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetMaxAdvance(nscoord &aAdvance)
{
  aAdvance = mMaxAdvance;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetFont(const nsFont*& aFont)
{
  aFont = mFont;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetLangGroup(nsIAtom** aLangGroup)
{
  if (!aLangGroup) {
    return NS_ERROR_NULL_POINTER;
  }

  *aLangGroup = mLangGroup;
  NS_IF_ADDREF(*aLangGroup);

  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetFontHandle(nsFontHandle &aHandle)
{
  aHandle = (nsFontHandle)mFontHandle;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetNormalLineHeight(nscoord &aHeight)
{
  aHeight = mEmHeight + mLeading;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetEmHeight(nscoord &aHeight)
{
  aHeight = mEmHeight;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetEmAscent(nscoord &aAscent)
{
  aAscent = mEmAscent;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetEmDescent(nscoord &aDescent)
{
  aDescent = mEmDescent;
  return NS_OK;
}

NS_IMETHODIMP  nsFontMetricsNanoX::GetMaxHeight(nscoord &aHeight)
{
  aHeight = mMaxHeight;
  return NS_OK;
}

nsresult nsFontMetricsNanoX::GetSpaceWidth(nscoord &aSpaceWidth)
{
  aSpaceWidth = mSpaceWidth;
    return NS_OK;
}

nsFontEnumeratorNanoX::nsFontEnumeratorNanoX()
{
  NS_INIT_REFCNT();
}

NS_IMPL_ISUPPORTS(nsFontEnumeratorNanoX, NS_GET_IID(nsIFontEnumerator));

NS_IMETHODIMP nsFontEnumeratorNanoX::EnumerateAllFonts(PRUint32 *aCount,
                                                       PRUnichar ***aResult)
{
  NS_ENSURE_ARG_POINTER(aResult);
  *aResult = nsnull;
  NS_ENSURE_ARG_POINTER(aCount);
  *aCount = 0;

  
  return EnumerateFonts(nsnull, nsnull, aCount, aResult);
}


static int 
CompareFontNames(const void* aArg1, const void* aArg2, void* aClosure) 
{ 
  const PRUnichar* str1 = *((const PRUnichar**) aArg1); 
  const PRUnichar* str2 = *((const PRUnichar**) aArg2); 
 
  // XXX add nsICollation stuff 
 
  return nsCRT::strcmp(str1, str2); 
} 

NS_IMETHODIMP nsFontEnumeratorNanoX::EnumerateFonts(const char *aLangGroup,
                                                    const char *aGeneric,
                                                    PRUint32 *aCount,
                                                    PRUnichar ***aResult)
{
	int j;
	PRUnichar **index, **array;

	NS_ENSURE_ARG_POINTER(aResult);
	*aResult = nsnull;
	NS_ENSURE_ARG_POINTER(aCount);
	*aCount = 0;
	
	if(gFontList == NULL)
		GrGetFontList(&gFontList, &gNumFonts);
	
	if(gNumFonts > -1) {
		array = (PRUnichar**)nsMemory::Alloc(gNumFonts * sizeof(PRUnichar*));
		index = array;
		j = 0;
		for(int i = 0; i < gNumFonts; i++) {
			nsCString *str = new nsCString(gFontList[i]->ttname);
			PRUnichar *pchar = str->ToNewUnicode();
			index[j] = pchar;
			j++;
			delete str;
		}
		
		NS_QuickSort(array, gNumFonts, sizeof(PRUnichar*), CompareFontNames,
				nsnull);
		
		*aResult = index;
		*aCount = j;
		return NS_OK;
	} else
		return NS_ERROR_FAILURE;
}

NS_IMETHODIMP nsFontEnumeratorNanoX::HaveFontFor(const char *aLangGroup,
                                                 PRBool *aResult)
{
  NS_ENSURE_ARG_POINTER(aResult);
  *aResult = PR_FALSE;
  NS_ENSURE_ARG_POINTER(aLangGroup);
  *aResult = PR_TRUE;
  return NS_OK;
}

NS_IMETHODIMP nsFontEnumeratorNanoX::UpdateFontList()
{
	if(gFontList)
		GrFreeFontList(&gFontList, gNumFonts);

	GrGetFontList(&gFontList, &gNumFonts);
	return NS_OK;
}
